/*
 * Camera.h
 *
 *  Created on: Jul 9, 2010
 *      Author: petrito1@cmp.felk.cvut.cz
 */

#ifndef NIFTI_CAMERA_CAMERA_H_
#define NIFTI_CAMERA_CAMERA_H_

#include <queue>
#include <set>

#include <dc1394/dc1394.h>
#include <opencv2/core/core.hpp>

namespace nifti {
namespace dc1394 {

/**
 * Object wrapper for libdc1394 cameras.
 * libdc1394 errors generally result into exceptions.
 *
 * Thread-safe for capturing from multiple threads. Ensures that frames are enqueued in
 * the same order as they were dequeued. Getters/setters thread-safety depends on libdc1394,
 * which is claimed to be thread-safe too.
 */
class Camera {

public:
  Camera(dc1394camera_t* camera);
  Camera();
  virtual ~Camera();

  /**
   * Setup operation mode, ISO speed, video mode, ROI etc.
   */
  void setup();

  /**
   * Returns the operation mode.
   */
  dc1394operation_mode_t getOperationMode();
  /**
   * Sets the operation mode.
   */
  void setOperationMode(dc1394operation_mode_t operationMode);

  /**
   * Returns the ISO speed.
   */
  dc1394speed_t getIsoSpeed();
  /**
   * Sets the ISO speed.
   */
  void setIsoSpeed(dc1394speed_t speed);

  /**
   * Returns the video mode.
   */
  dc1394video_mode_t getVideoMode();
  /**
   * Sets the video mode.
   */
  void setVideoMode(dc1394video_mode_t videoMode);

  /**
   * Returns Format 7 colod coding.
   */
  dc1394color_coding_t getFormat7ColorCoding();
  /**
   * Return Format 7 color filter.
   */
  dc1394color_filter_t getFormat7ColorFilter();
  /**
   * Returns Format 7 frame interval.
   */
  float getFormat7FrameInterval();
  /**
   * Returns number of packets per frame for Format 7.
   */
  uint32_t getFormat7PacketsPerFrame();
  /**
   * Returns current Format 7 packet size.
   */
  uint32_t getFormat7PacketSize();
  /**
   * Returns recommended Format 7 packet size.
   */
  uint32_t getFormat7RecommendedPacketSize();

  /**
   * Returns Format 7 ROI.
   */
  void getFormat7Roi(dc1394video_mode_t videoMode, dc1394color_coding_t* colorCoding, uint32_t* packetSize,
      uint32_t* left, uint32_t* top, uint32_t* width, uint32_t* height);
  /**
   * Sets Format 7 ROI.
   */
  virtual void setFormat7Roi(dc1394video_mode_t videoMode, dc1394color_coding_t colorCoding, int32_t packetSize, int32_t left,
      int32_t top, int32_t width, int32_t height);

  /**
   * Initializes DMA buffers.
   */
  void setupCapture();

  /**
   * Frees initialized DMS buffers.
   */
  void stopCapture();

  /**
   * Flashes all frames, does not wait for frames.
   */
  void flushFrames();

  /**
   * Drops the specified number of frames, waits for frames.
   */
  void dropFrames(int n);

  /**
   * Starts video transmission using initialized buffers.
   */
  void startVideoTransmission();
  /**
   * Stops video transmission.
   */
  void stopVideoTransmission();

  /**
   * Enables one-shot video. Then a single frame may be captured by captureColorMat.
   *
   * Not in effect when video transmission is enabled.
   * User may want to flush unwanted frames before.
   */
  void startOneShot();

  /**
   * Disables one-shot video.
   */
  void stopOneShot();

  /**
   * Debayers the in frame, fills the out frame with color image.
   * Output color-coding is DC1394_COLOR_CODING_RGB8.
   * Do not pre-allocate the image buffer of the out frame.
   *
   * @param in pointer to the source frame, not null.
   * @param out pointer to the output frame, not null.
   */
  void debayerFrame(dc1394video_frame_t* in, dc1394video_frame_t* out);

  /**
   * Captures a frame.
   *
   * @param bgrMat color image matrix
   * @param timestamp frame timestamp
   * @param wait 'wait for frame' flag
   */
  void captureColorMat(cv::Mat& bgrMat, uint64_t& timestamp, bool wait = true);

  /**
   * Powers on the camera.
   */
  void powerOn();
  /**
   * Powers off the camera.
   */
  void powerOff();

  /**
   * Resets the camera to factory default settings.
   */
  void reset();

  /**
   * Resets the bus to which this camera is attached.
   */
  void resetBus();

  /**
   * Releases all previously allocated ISO channels and bandwidth.
   */
  void releaseAll();

  dc1394bayer_method_t getBayerMethod();
  void setBayerMethod(dc1394bayer_method_t method);

  /**
   * Dequeues a frame from DMA buffer and returns its pointer. Thread-safe.
   * After processing, the frame must be returned back to the buffer
   * by calling enqueueFrameLater and enqueueFrames.
   *
   * @param capturePolicy a capture policy to use (wait/poll).
   * @returns pointer to a frame from DMA buffer.
   */
  dc1394video_frame_t* dequeueFrame(const dc1394capture_policy_t capturePolicy);

  bool isFrameCorrupt(dc1394video_frame_t* frame);

  /**
   * Marks the frame as ready for enqueuing. Thread-safe.
   * @param frame pointer to the frame.
   */
  void enqueueFrameLater(const dc1394video_frame_t* frame);

  /**
   * Enqueue all frames marked as ready, e.g. using enqueueFrameLater. Thread-safe.
   */
  void enqueueFrames();

  /**
   * Returns value of the camera register.
   */
  uint32_t getRegister(uint64_t offset);

  /**
   * Sets value of the camera register.
   */
  void setRegister(uint64_t offset, uint32_t value);

  uint32_t getControlRegister(uint64_t offset);
  void setControlRegister(uint64_t offset, uint32_t value);

  /**
   * Captures the dequeue/enqueue lock.
   */
  void lockCapture();

  /**
   * Releases the dequeue/enqueue lock.
   */
  void unlockCapture();

  /**
   * Returns true if the feature is present, false otherwise.
   */
  bool isFeaturePresent(dc1394feature_t feat);

  /**
   * Returns true if the feature is readable, false otherwise.
   */
  bool isFeatureReadable(dc1394feature_t feat);

  /**
   * Returns true if the feature is switchable, false otherwise.
   * False means that feature is always on.
   */
  bool isFeatureSwitchable(dc1394feature_t feat);

  /**
   * Returns true if the feature is powered on, false otherwise.
   */
  bool isFeaturePowerOn(dc1394feature_t feat);

  /**
   * Returns the feature mode.
   */
  dc1394feature_mode_t getFeatureMode(dc1394feature_t feat);

  /**
   * Returns true if feature absolute control is available, false otherwise.
   */
  bool isFeatureAbsoluteControlAvailable(dc1394feature_t feat);

  /**
   * Returns true if feature absolute control is on, false otherwise.
   */
  bool isFeatureAbsoluteControlOn(dc1394feature_t feat);

  /**
   * Returns the minimum value of the feature.
   */
  uint32_t getFeatureMinBoundary(dc1394feature_t feat);

  /**
   * Returns the maximum value of the feature.
   */
  uint32_t getFeatureMaxBoundary(dc1394feature_t feat);

  /**
   * Returns the current value of the feature.
   */
  uint32_t getFeatureValue(dc1394feature_t feat);

  /**
   * Returns the minimum absolute value of the feature.
   */
  float getFeatureMinAbsoluteBoundary(dc1394feature_t feat);

  /**
   * Returns the maximum absolute value of the feature.
   */
  float getFeatureMaxAbsoluteBoundary(dc1394feature_t feat);

  /**
   * Returns the current absolute value of the feature.
   */
  float getFeatureAbsoluteValue(dc1394feature_t feat);

  void setFeaturePowerOn(dc1394feature_t feat);

  void setFeaturePowerOff(dc1394feature_t feat);

  /**
   * Sets feature mode.
   *
   * Possible values (for convenience, see dc1394feature_mode_t definition):
   * DC1394_FEATURE_MODE_MANUAL
   * DC1394_FEATURE_MODE_AUTO
   * DC1394_FEATURE_MODE_ONE_PUSH_AUTO
   */
  void setFeaturetMode(dc1394feature_t feat, dc1394feature_mode_t mode);

  void setFeatureAbsoluteControlOn(dc1394feature_t feat);

  void setFeatureAbsoluteControlOff(dc1394feature_t feat);

  /**
   * Sets feature value (should be within the boundaries).
   */
  void setFeatureValue(dc1394feature_t feat, uint32_t value);

  /**
   * Sets feature absolute value (should be within absolute boundaries & absolute control should be enabled).
   */
  void setFeatureAbsoluteValue(dc1394feature_t feat, float value);

  /**
   * Returns the value of the blue (in RGB) (U in YUV) channel.
   */
  uint32_t getWhiteBalanceBlue();

  /**
   * Returns the value of the red (in RGB) (V in YUV) channel.
   */
  uint32_t getWhiteBalanceRed();

  /**
   * Sets the value of the blue (in RGB) (U in YUV) and red (in RGB) (V in YUV) channel.
   */
  void setWhiteBalance(uint32_t u_b_value, uint32_t v_r_value);

  /**
   * Sets the value of the blue (in RGB) (U in YUV) channel.
   */
  void setWhiteBalanceBlue(uint32_t u_b_value);

  /**
   * Sets the value of the red (in RGB) (V in YUV) channel.
   */
  void setWhiteBalanceRed(uint32_t v_r_value);

private:
  /**
   * The synchronization mutex for dequeuing and enqueuing image buffers.
   * Buffers should be enqueued in the same order as they were dequeued.
   */
  pthread_mutex_t captureMutex;

  /**
   * A dc1394 camera to which dc1394-related operations are delegated.
   */
  dc1394camera_t* camera;

  /**
   * Dequeued frames, maintains correct order.
   */
  std::queue<dc1394video_frame_t*> dequeuedFrames;

  /**
   * Processed frames to enqueue in correct order.
   */
  std::set<dc1394video_frame_t*> framesToEnqueue;

  /**
   * Bayer method to use in capturing.
   */
  dc1394bayer_method_t bayerMethod;

};

}
}

#endif /* NIFTI_CAMERA_CAMERA_H_ */
