/*
 * LadybugVirtualCamera.h
 *
 *  Created on: Aug 5, 2010
 *      Author: petrito1@cmp.felk.cvut.cz
 */

#ifndef NIFTI_LADYBUG_LADYBUGVIRTUALCAMERA_H_
#define NIFTI_LADYBUG_LADYBUGVIRTUALCAMERA_H_

#include <GL/glew.h>
#include <GL/gl.h>

#include <opencv2/core/core.hpp>

namespace nifti {
namespace ladybug {

/**
 * @brief Virtual camera for Ladybug 3.
 *
 * Renders a virtual single-lens camera view, using images captured from Ladybug 3.
 * A camera can be initialized using a mesh data and alphamasks generated by Ladybug
 * SDK.
 */
class LadybugVirtualCamera {
private:
  static const int NUM_CAMERAS = 6;

  /**
   * Rendered viewport width.
   */
  int viewportWidth;
  /**
   * Rendered viewport height.
   */
  int viewportHeight;

  /**
   * Horizontal pan, i.e. azimuth.
   */
  double pan;
  /**
   * Vertical tilt, i.e. elevation, altitude.
   */
  double tilt;
  /**
   * Horizontal field of view, in degrees.
   */
  double horizontalFov;
  /**
   * Vertical field of view, in degrees.
   */
  double verticalFov;

  /**
   * Number of columns in the mesh.
   */
  int meshCols;
  /**
   * Number of rows in the mesh.
   */
  int meshRows;
  /**
   * Mesh buffer with points as (x, y, z) triples.
   * NB: Allocated with new, e.g. before reading the mesh file.
   * Must be freed by delete.
   */
  double* mesh[NUM_CAMERAS];

  /**
   * Is alphamask available and ready to use?
   */
  bool alphamaskAvailable;
  /**
   * Widths of individual alphamasks.
   */
  int alphaMaskWidths[NUM_CAMERAS];
  /**
   * Heights of individual alphamasks.
   */
  int alphaMaskHeights[NUM_CAMERAS];
  /**
   * Pointers to alphamask buffers.
   * NB: Allocated with malloc, e.g. when reading pgm files.
   * Must be freed by free.
   */
  unsigned char* alphamasks[NUM_CAMERAS];

  /**
   * Image widths.
   */
  int imageWidths[NUM_CAMERAS];
  /**
   * Image heights.
   */
  int imageHeights[NUM_CAMERAS];

  /**
   * Assigned GL texture IDs.
   */
  GLuint glTextures[NUM_CAMERAS];
  /**
   * Image texture widths.
   */
  int glTextureWidths[NUM_CAMERAS];
  /**
   * Image texture heights.
   */
  int glTextureHeights[NUM_CAMERAS];
  /**
   * Properly initialized?
   */
  bool initialized;
  /**
   * Is off-screen rendering enabled, e.g. using framebuffers?
   */
  bool offScreenRendering;
  /**
   * Frame buffer.
   */
  GLuint frameBuffer;
  /**
   * Color render buffer.
   */
  GLuint colorRenderBuffer;
  /**
   * Depth render buffer.
   */
  GLuint depthRenderBuffer;

  /**
   * Returns minimum power of 2 greater or equal to n.
   */
  int getMinimumPowerOfTwo(int n);

  /**
   * Updates GL frame and render buffers according to current viewport size.
   */
  void updateBuffers();

  /**
   * Frees GL frame and render buffers.
   */
  void deleteBuffers();
  /**
   * Initializes GL textures.
   */
  void initializeTextures();
  /**
   * Initializes GL framebuffers.
   */
  void initializeFramebuffers();

public:
  LadybugVirtualCamera();
  virtual ~LadybugVirtualCamera();

  /**
   * Initializes GL context.
   *
   * @param offScreenRendering Enable off-screen rendering?
   */
  void initialize(bool offScreenRendering);

  /**
   * Loads the mesh from file, as saved with Ladybug SDK.
   */
  void loadMesh(const std::string& meshFile);
  /**
   * Clears current mesh and frees associated resources.
   */
  void clearMesh();

  /**
   * Loads alphamasks from files, as saved with Ladybug SDK.
   * Sets alphamaskAvailable to true if alphamasks loads successfully.
   */
  void loadAlphamask(const std::string& alphamaskFilePrefix);

  /**
   * Clears current alphamasks and frees associated resources.
   * Sets alphamaskAvailable to false.
   */
  void clearAlphamask();

  /**
   * Sets the i-th image.
   */
  void setImage(const int& i, const cv::Mat& image);
  /**
   * Sets all images.
   */
  void setImages(const cv::Mat(&images)[NUM_CAMERAS]);
  /**
   * Sets all images from the composite.
   */
  void setImages(const cv::Mat& composite);

  /**
   * Renders current view to the image matrix.
   */
  void renderView(cv::Mat& image);

  void getViewport(int* viewportWidth, int* viewportHeight);
  void setViewport(int viewportWidth, int viewportHeight);

  double getPan();
  void setPan(double pan);
  double getTilt();
  void setTilt(double tilt);
  double getHorizontalFov();
  void setHorizontalFov(double horizontalFov);
  double getVerticalFov();
  void setVerticalFov(double verticalFov);

  // Made public so the they can be used as GLUT callbacks:
  /**
   * Updates viewport transform based on current viewport width and height.
   */
  void updateViewport();

  /**
   * Updates projection based on current horizontal and vertical field of view.
   */
  void updateProjection();

  /**
   * Update model view transform based on current pan and tilt angles.
   */
  void updateModelView();

  /**
   * Renders the scene.
   */
  void render();

  /**
   * Clears frame and render buffers.
   */
  void clearBuffers();

  /**
   * Is initialized.
   */
  bool isInitialized();

};

}
}

#endif /* NIFTI_LADYBUG_LADYBUGVIRTUALCAMERA_H_ */
