
#include <ros/ros.h>
#include <cv_bridge/cv_bridge.h>
#include <image_transport/image_transport.h>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/opencv.hpp>
#include <sensor_msgs/image_encodings.h>

#include "omnicamera/LookupStitcher.h"

namespace {
int inWidth, inHeight;
int outWidth, outHeight;
bool alpha;
int radius;
string lutPath;

omnicamera::LookupStitcher stitcher;
image_transport::Publisher outPub;
ros::Publisher camPub;
}

void imageCallback(const sensor_msgs::ImageConstPtr& msg) {

  cv_bridge::CvImageConstPtr cvPtr;
  try {
    cvPtr = cv_bridge::toCvShare(msg);
    if (cvPtr->encoding.compare(sensor_msgs::image_encodings::BGR8)) {
      ROS_ERROR("Encoding not supported: %s.", cvPtr->encoding.data());
      return;
    }
  } catch (cv_bridge::Exception& e) {
    ROS_ERROR("cv_bridge exception: %s", e.what());
    return;
  }

  if (cvPtr->image.rows != inHeight || cvPtr->image.cols != inWidth) {
    ROS_ERROR("Unexpected image dimensions: %dx%d.", cvPtr->image.cols, cvPtr->image.rows);
  }

  Mat image;
  stitcher.stitchImage(cvPtr->image, image);

  cv_bridge::CvImage out;
  out.encoding = sensor_msgs::image_encodings::BGR8;
  out.header = msg->header;
  out.image = image;
  outPub.publish(out.toImageMsg());

  sensor_msgs::CameraInfo camMsg;
  camMsg.header = out.header;
  camMsg.height = out.image.rows;
  camMsg.width = out.image.cols;
  camPub.publish(camMsg);
}

int main(int argc, char ** argv) {

  ros::init(argc, argv, "img2pano");
  ros::NodeHandle nh;

  // Construct a node handle with private namespace to get private parameters...
  ros::NodeHandle pnh("~");
  pnh.param("inWidth", inWidth, 1616);
  pnh.param("inHeight", inHeight, 1232);
  ROS_INFO("Input image size: %dx%d.", inWidth, inHeight);

  pnh.param("outWidth", outWidth, 960);
  pnh.param("outHeight", outHeight, 480);
  ROS_INFO("Output image size: %dx%d.", outWidth, outHeight);

  pnh.param("alpha", alpha, true);
  ROS_INFO("Alpha blending: %s.", alpha ? "enabled" : "disabled");

  pnh.param("radius", radius, 20);
  ROS_INFO("Sphere radius: %d.", radius);

  pnh.param("lutPath", lutPath, string("LUT_img_%dx%d_pano_%dx%dr%d_%s%d.pgm"));
  ROS_INFO("LUT path template: %s.", lutPath.data());

  stitcher.clearLookupTables();
  for (int iCam = 1; iCam <= 6; iCam++) {
    int len = 1024;
    char xPath[len], yPath[len], aPath[len];
    snprintf(xPath, len, lutPath.data(), inWidth, inHeight, outWidth, outHeight, radius, "x", iCam);
    snprintf(yPath, len, lutPath.data(), inWidth, inHeight, outWidth, outHeight, radius, "y", iCam);
    snprintf(aPath, len, lutPath.data(), inWidth, inHeight, outWidth, outHeight, radius, "a", iCam);
    ROS_INFO("Camera %d look-up tables: %s, %s, %s.", iCam, xPath, yPath, aPath);
    stitcher.addLookupTables(string(xPath), string(yPath), string(aPath));
  }
  
  image_transport::ImageTransport it(nh);
  image_transport::Subscriber sub = it.subscribe("in", 1, imageCallback);
  outPub = it.advertise("out", 1);
  camPub = nh.advertise<sensor_msgs::CameraInfo>("camera_info", 1);

  ros::spin();

  return 0;
}

