/*
 * ladybug_test.cpp
 *
 *  Created on: Aug 24, 2010
 *      Author: petrito1@cmp.felk.cvut.cz
 *
 * Executable for tracking corrupt images.
 * Corrupt images are detected using mean difference between two successive images.
 * The reason is most probably a HW configuration:
 * http://www.ptgrey.com/support/kb/index.asp?a=4&q=188
 */

#include <GL/glew.h>
#include <GL/glut.h>

#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>

#include <signal.h>
#include <sys/param.h>
#include <sys/times.h>

#include "CameraException.h"
#include "Camera.h"
#include "CameraContext.h"

#include "LadybugVirtualCamera.h"

using std::string;

using cv::Mat;

using nifti::dc1394::Camera;
using nifti::dc1394::CameraContext;
using nifti::dc1394::CameraException;

using nifti::ladybug::LadybugVirtualCamera;

void sigintHandler(int sig);
void printDuration(char* message, tms t1, long r1);

namespace {
bool halt = false;
}

int main(int argc, char **argv) {

  if (argc < 3) {
    return -1;
  }

  signal(SIGINT, sigintHandler);
  glutInit(&argc, argv);

  CameraContext context;

  Camera camera = context.getCamera();
  //  camera.resetBus();
  //  camera.reset();
//  camera.setup();

  printf("Format 7 color coding: %d.\n", camera.getFormat7ColorCoding());
  printf("Format 7 color filter: %d.\n", camera.getFormat7ColorFilter());
  printf("Format 7 frame interval: %f.\n", camera.getFormat7FrameInterval());
  printf("Format 7 packets per frame: %u.\n", camera.getFormat7PacketsPerFrame());
  printf("Format 7 packet size: %u.\n", camera.getFormat7PacketSize());
  printf("Format 7 recommended packet size: %u.\n", camera.getFormat7RecommendedPacketSize());

  camera.setupCapture();
  camera.startVideoTransmission();

  LadybugVirtualCamera ladybug;
  ladybug.initialize(true);
  ladybug.loadMesh(argv[1]);
  ladybug.loadAlphamask(argv[2]);
  ladybug.setViewport(1200, 800);
  ladybug.setHorizontalFov(120);
  ladybug.setVerticalFov(80);

  struct tms t1;
  long r1;

  int iFrame = 0;
  Mat frame1, frame2;
  while (!halt) {
    r1 = times(&t1);
    camera.captureColorMat(frame2, true);
    printDuration("Frame capture:     ", t1, r1);

    // Compare two following images for differences.
    if (frame1.rows > 0 && frame1.cols > 0) {

      r1 = times(&t1);
      double diff = std::abs(cv::mean(frame1 - frame2).val[0]);
      printf("Absolute mean difference: %lf\n", diff);
      printDuration("Calculating difference:  ", t1, r1);

      if (diff > 25.0) {
        r1 = times(&t1);
        char imagePath[128];
        sprintf(imagePath, "image_%02d_composite.jpg", iFrame);
        cv::imwrite(string(imagePath), frame2);
        printDuration("Writing composite: ", t1, r1);
      }
    }

    //    r1 = times(&t1);
    //    vcam.setImages(composite);
    //    printDuration("Setting images:    ", t1, r1);

    //    Mat view;
    //    r1 = times(&t1);
    //    vcam.renderView(view);
    //    printDuration("Rendering view:    ", t1, r1);

    //    r1 = times(&t1);
    //    sprintf(imagePath, "image_%02d_view.jpg", iFrame++);
    //    cv::imwrite(string(imagePath), view);
    //    printDuration("Writing view:      ", t1, r1);

    iFrame++;
    frame1 = frame2;
  }

  camera.stopVideoTransmission();
  camera.stopCapture();
  camera.releaseAll();
}

void sigintHandler(int sig) {
  printf("Signal %d received. Halting...\n", sig);
  halt = true;
}

void printDuration(char* message, tms t1, long r1) {
  struct tms t2;
  long r2;
  r2 = times(&t2);

  printf("%s", message);
  printf("user = %1.2f, ", ((float) (t2.tms_utime - t1.tms_utime)) / HZ);
  printf("system = %1.2f, ", ((float) (t2.tms_stime - t1.tms_stime)) / HZ);
  printf("real = %1.2f\n", ((float) (r2 - r1)) / HZ);
}
